"""
`mfcrypt.py`

Python cryptographic utilities for Musicfox applications.

"""
from Cryptodome.Protocol import KDF
from Cryptodome.Cipher import AES
from Cryptodome.Util.Padding import pad, unpad
from base64 import b64decode, b64encode
import json
import hmac
from hashlib import sha256
import codecs
from typing import Union


class Tuple(tuple):
    pass


def create_bytes_key(
    passphrase: str, salt: str, iterations: int = 128
) -> bytes:
    """
    Create the PBKDF2 key given the `passphrase`, `salt`, and optionally adjustable
    `iterations`.

    > :info: _Note: `iterations` should be set as high as budget and user-experience allow for best security._

    Returns the resulting bytes from the `PBKDF2` hash algorithm.
    """
    return KDF.PBKDF2(passphrase, salt.encode("utf-8"), 48, iterations)


def _convert_to_bytes(*args):
    """Attempt to convert the given arguments to bytes. Though tested with
    strings, an individual arguments of *args, serializable to JSON, should
    work.
    """
    results = []
    for arg in args:
        if isinstance(arg, str):  # skip
            results.append(codecs.encode(arg))
        elif isinstance(arg, bytes):
            results.append(arg)
        elif isinstance(arg, dict):
            results.append(json.dumps(arg).encode("utf-8"))
        elif isinstance(arg, list):
            results.append(json.dumps(arg).encode("utf-8"))
        else:
            print(f"{arg} is {type(arg)} and will not be in the results array")
    return results


def add_signature(secret_key: str, data: str) -> str:
    """
    Generate an HMAC digest using the given secret key and data string.

    >_Note: This produces a hex digest string for use in message authentication._
    """
    digest = hmac.new(
        secret_key.encode("utf-8"), data.encode("utf-8"), sha256
    ).hexdigest()
    return digest


def check_signature(
    hmac_digest: Union[bytes, str],
    secret_key: str,
    data: Union[bytes, str],
) -> bool:
    """
    Check that the given hmac_digest computes to that of the one
    generated from the secret_key and data combination.

    Specifically, use the `compare_digest` method from the [`hmac`](https://docs.python.org/3/library/hmac.html) standard library module.
    """
    if type(hmac_digest) != type(data):
        raise TypeError(
            f"Input types must be the same. hmac_digest is type {type(hmac_digest)} and data is type {type(data)}."
        )

    if isinstance(data, bytes) and isinstance(hmac_digest, bytes):
        computed = add_signature(
            secret_key=secret_key, data=data.decode("utf-8")
        )
        return hmac.compare_digest(hmac_digest, computed.encode("utf-8"))
    computed = add_signature(secret_key=secret_key, data=data)
    hmac_bytes, computed_bytes = _convert_to_bytes(hmac_digest, computed)
    return hmac.compare_digest(hmac_bytes, computed_bytes)


def _extract_iv_and_key(secret_key_bytes: bytes, block_size: int = 16):
    """
    INTERNAL

    Extract the initialization vector and the key as represented in bytes.
    """
    iv, key = (
        secret_key_bytes[:block_size],
        secret_key_bytes[block_size : (block_size + 2 * block_size)],
    )
    return iv, key


def decrypt(data, secret: str, salt: str, type_hint: str = "string"):
    """
    Decrypt the given `data` (bytes string) via the given `secret_key_bytes`
    which includes the `iv` and `key` values.

    This function unpads the data with a default block size of 16 and
    returns the decrypted data as as Python 3 string (unicode).
    """
    if not isinstance(data, bytes) and isinstance(data, str):
        # data is str, convert to byte string
        data = codecs.encode(data)
    encdata, digest = data.split(sep=b"::")

    if not check_signature(digest, secret_key=secret, data=encdata):
        computed = add_signature(secret, encdata.decode("utf-8"))
        msg = f"data: {data}\nencdata: {encdata}\ndigest: {digest}\ncomputed: {computed}\n"
        raise ValueError(
            f"Hash comparison failed. This has automatically been reported to the United States Federal Bureau of Investigation as a potential cyber crime.\n{msg}"
        )
    data = b64decode(encdata)
    secret_key_bytes = create_bytes_key(secret, salt)
    iv, key = _extract_iv_and_key(secret_key_bytes)
    cipher = AES.new(key, AES.MODE_CBC, iv)
    text = unpad(cipher.decrypt(data), 16)
    if type_hint != "string":
        return json.loads(text.decode("utf-8"))
    return text.decode("utf-8")


def encrypt(data, secret: str, salt: str, type_hint: str = "string") -> str:
    """
    Encrypt the given `data` (bytes string) via the given `secret_key_bytes`, which includes the
    `iv` and `key` values.

    This function pads the data prior to encryption with a default block size of 16 and returns
    the base64-encoded encrypted data as a byte string.
    """
    if isinstance(data, dict):
        data = json.dumps(data)
    if isinstance(data, str):
        data = data.encode("utf-8")
    if not isinstance(data, (bytes, bytearray)):
        raise TypeError("Data must be a string, bytes, bytearray, or a dict.")
    secret_key_bytes = create_bytes_key(secret, salt)
    iv, key = _extract_iv_and_key(secret_key_bytes)
    cipher = AES.new(key, AES.MODE_CBC, iv)
    enctext = cipher.encrypt(pad(data, 16))
    b64enctext = b64encode(enctext)
    computed_hash = add_signature(
        secret_key=secret, data=b64enctext.decode("utf-8")
    ).encode("utf-8")
    return (b64enctext + b"::" + computed_hash).decode("utf-8")



import mfcrypt
import codecs


def test_convert_to_bytes():
    _convert_to_bytes = getattr(mfcrypt, "_convert_to_bytes")
    test1, test2 = "testing", b"testingbytes"
    results = _convert_to_bytes(test1, test2)
    assert isinstance(results, list)
    for item in results:
        assert isinstance(item, bytes)

    big_test = ["testing"] * 1000
    big_test[800] = codecs.encode(big_test[800])  # throw in a funky one
    results = _convert_to_bytes(*big_test)
    for item in results:
        assert isinstance(item, bytes)

    # try dict + list
    test1, test2 = ["testing"], dict(testing="testingbytes")
    results = _convert_to_bytes(test1, test2)
    assert isinstance(results, list)
    for item in results:
        assert isinstance(item, bytes)
    # tuple won't/shouldn't work but not throw
    testtup = ("test", "testing", "tested")
    results = _convert_to_bytes(testtup)
    assert isinstance(results, list)
    assert not len(results)


def test_add_signature_hmac_sha256(crypt_data):
    test_hash = mfcrypt.add_signature(
        crypt_data.passphrase, crypt_data.encrypted
    )
    assert isinstance(test_hash, str)
    assert test_hash != mfcrypt.add_signature(
        "not the right key", crypt_data.encrypted
    )


def test_check_signature_hmac_sha256(crypt_data):
    test_hash = mfcrypt.add_signature(
        crypt_data.passphrase, crypt_data.plaintext
    )
    mfcrypt.check_signature(
        test_hash, crypt_data.passphrase, crypt_data.plaintext
    )


def test_create_bytes_key(crypt_data):
    secret_key_bytes = mfcrypt.create_bytes_key(
        crypt_data.passphrase, crypt_data.salt
    )
    assert isinstance(secret_key_bytes, bytes)


def test_encrypt_string(crypt_data):
    encrypted = mfcrypt.encrypt(
        crypt_data.plaintext, crypt_data.passphrase, crypt_data.salt
    )
    assert isinstance(encrypted, str)
    assert encrypted != crypt_data.plaintext
    assert encrypted == crypt_data.encrypted_payload
    assert len(encrypted)


def test_encrypt_dict(crypt_data):
    encrypted = mfcrypt.encrypt(
        crypt_data.obj_data, crypt_data.passphrase, crypt_data.salt
    )
    assert isinstance(encrypted, str)
    assert encrypted != crypt_data.obj_data
    assert encrypted == crypt_data.encrypted_object_data
    assert len(encrypted)


def test_encrypt_bytearray(crypt_data):
    encoded = bytearray(crypt_data.plaintext.encode("utf-8"))
    try:
        encrypted = mfcrypt.encrypt(
            encoded, crypt_data.passphrase, crypt_data.salt
        )
        assert isinstance(encrypted, str)
        assert encrypted != crypt_data.plaintext.encode("utf-8")
        assert encrypted == crypt_data.encrypted
        assert len(encrypted)
    except TypeError as terr:
        assert (
            "Data must be a string, bytes, bytearray, or a dict." in f"{terr}"
        )


def test_encrypt_bytearray(crypt_data):
    class WillNotWork:
        def __init__(self):
            self.try_me = False

        def __repr__(self):
            return f"try_me: {self.try_me}"

    try:
        wnw = WillNotWork()
        encrypted = mfcrypt.encrypt(
            wnw, crypt_data.passphrase, crypt_data.salt
        )
        assert isinstance(encrypted, bytes)
        assert encrypted != crypt_data.plaintext.encode("utf-8")
        assert encrypted == crypt_data.encrypted
        assert len(encrypted)
    except TypeError as terr:
        assert (
            "Data must be a string, bytes, bytearray, or a dict." in f"{terr}"
        )


def test_decrypt_string(crypt_data):
    decrypted = mfcrypt.decrypt(
        crypt_data.encrypted_payload,
        crypt_data.passphrase,
        crypt_data.salt,
        type_hint="string",
    )
    assert isinstance(decrypted, str)
    assert decrypted == crypt_data.plaintext


def test_decrypt_string_bad_hash(crypt_data):
    try:
        decrypted = mfcrypt.decrypt(
            b"random-badness::notahash",
            crypt_data.passphrase,
            crypt_data.salt,
            type_hint="string",
        )
    except ValueError as verr:
        assert "Hash comparison failed" in f"{verr}"


def test_decrypt_dict(crypt_data):
    decrypted = mfcrypt.decrypt(
        crypt_data.encrypted_object_data,
        crypt_data.passphrase,
        crypt_data.salt,
        type_hint="dict",
    )
    assert isinstance(decrypted, dict)
    assert decrypted == crypt_data.obj_data



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Note: To use the 'upload' functionality of this file, you must:
#   $ pipenv install twine --dev

import io
import os
import sys
from shutil import rmtree
import mfcrypt
from setuptools import find_packages, setup, Command
from setuptools.command.egg_info import egg_info

# Package meta-data.
NAME = "mfcrypt"
DESCRIPTION = "Cryptographic utilities for Musicfox Python applications."
URL = "https://github.com/musicfox/cryptography/"
EMAIL = "dev@musicfox.io"
LICENSE = "MIT"
AUTHOR = "Jason R. Stevens, CFA on behalf of Musicfox, Inc."
REQUIRES_PYTHON = ">=3.9"
VERSION = mfcrypt.__version__
GEMFURY = None  # f'twine upload dist/* --repository-url https://push.fury.io/musicfox -u {os.environ.get("TWINE_TOKEN")} -p ""'

# What packages are required for this module to be executed?
REQUIRED = [
    "pycryptodomex",
]

# What packages are optional?
EXTRAS = {
    # 'fancy feature': ['django'],
}

# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(here, "README.md"), encoding="utf-8") as f:
        long_description = "\n" + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

# Load the package's __version__.py module as a dictionary.
about = {}
if not VERSION:
    project_slug = NAME.lower().replace("-", "_").replace(" ", "_")
    with open(os.path.join(here, project_slug, "__version__.py")) as f:
        exec(f.read(), about)
else:
    about["__version__"] = VERSION


class UploadCommand(Command):
    """Support setup.py upload."""

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print("\033[1m{0}\033[0m".format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds…")
            rmtree(os.path.join(here, "dist"))
        except OSError:
            pass

        self.status("Building Source and Wheel distribution…")
        os.system("{0} setup.py sdist bdist_wheel ".format(sys.executable))

        if GEMFURY:
            self.status("Uploading the package to Gemfury via Twine...")
            os.system(GEMFURY)

        else:
            self.status("Uploading the package to PyPI via Twine…")

            os.system("twine upload dist/*")

        sys.exit()


class egg_info_ex(egg_info):
    """Includes license file into `.egg-info` folder."""

    def run(self):
        # don't duplicate license into `.egg-info` when building a distribution
        if not self.distribution.have_run.get("install", True):
            # `install` command is in progress, copy license
            self.mkpath(self.egg_info)
            self.copy_file("LICENSE", self.egg_info)

        egg_info.run(self)


# Where the magic happens:
setup(
    name=NAME,
    version=about["__version__"],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    # If your package is a single module, use this instead of 'packages':
    # py_modules=['mfcrypt'],
    packages=find_packages(
        exclude=["tests", "*.tests", "*.tests.*", "tests.*", "node_modules"]
    ),
    # entry_points={
    #     'console_scripts': ['mycli=mymodule:cli'],
    # },
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    include_package_data=True,
    license=LICENSE,
    license_files=("LICENSE",),
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
    ],
    # $ setup.py publish support.
    cmdclass={"upload": UploadCommand, "egg_info": egg_info_ex},
)



"""test_setup.py 

# Test config & setup items

This file outlines `pytest` functions to cover important project 
assumptions related to configuration and setup items. 

## Usage 

Running a `pytest` script is simple. Here's how:
```python 
python -m pytest -vv --cov --cov-branch test/test_setup.py
```
"""
import mfcrypt


def test_package_version():
    """Test that the package version is a string and has at least two periods"""
    assert isinstance(mfcrypt.__version__, str)
